# -*- coding: utf-8 -*-
# vim:fenc=utf-8


import os
from schema import Schema, Optional, Or


DEFAULT_CONFIG = {'algorithm': {}, 'manager': {}}

DEFAULT_LOGPATH = '/tmp'

DEFAULT_HOME = '/'.join([os.environ.get('HOME', '/'), '.intuition'])

# https://github.com/halst/schema
# NOTE Even better ? https://github.com/j2labs/schematics
# TODO More strict validation
CONFIG_SCHEMA = Schema({
    'universe': basestring,
    'index': object,
    Optional('_id'): object,
    # Optional('id'): Use(basestring, error='invalid identity'),
    Optional('id'): basestring,
    Optional('live'): bool,
    Optional('frequency'): basestring,
    Optional('_id'): object,
    Optional('__v'): object,
    'modules': {
        'algorithm': basestring,
        # TODO It will be at least one
        Optional('backtest'): basestring,
        Optional('live'): basestring,
        Optional('manager'): Or(basestring, None)
    }
})


PANDAS_FREQ = {
    'daily': 'D',
    'hourly': 'H',
    'minutely': 'Min'
}



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition Context api
  ---------------------

  Context factory api.

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import abc
import pytz
import datetime as dt
import dna.logging
import intuition.utils


def parse_storage(storage):
    tmp_uri = storage.split('?')
    path = tmp_uri[0].split('/')[1:]
    uri = tmp_uri[0].split('/')[0]

    params = {}
    if len(tmp_uri) > 1:
        tmp_params = tmp_uri[1].split('&')
        for item in tmp_params:
            if item.find('=') > 0:
                k, v = item.split('=')
                params[k] = v
            else:
                params[item] = True

    return {
        'uri': uri,
        'path': path,
        'params': params
    }


class ContextFactory():
    '''
    Contexts loader gives to intuition everything it needs to know about user
    configuration. It also provides some methods to make the setup process
    easier (like the interpreting the shortcut 'nasdaq,6' into 6 random symbols
    from nasdaq exchange).
    '''

    __metaclass__ = abc.ABCMeta

    def __init__(self, storage):
        self.log = dna.logging.logger(__name__)
        self.initialize(parse_storage(storage))

    def initialize(self, storage):
        ''' Users should overwrite this method '''
        pass

    @abc.abstractmethod
    def load(self, uri, path, params):
        ''' Users should overwrite this method '''
        pass

    def _normalize_dates(self, context):
        '''
        Build a timeline from the start, end and frequency infos
        '''
        if 'start' in context:
            if isinstance(context['start'], dt.date):
                context['start'] = dt.date.strftime(
                    context['start'], format='%Y-%m-%d')
        if 'end' in context:
            if isinstance(context['end'], dt.date):
                context['end'] = dt.date.strftime(
                    context['end'], format='%Y-%m-%d')
        context['frequency'] = context.get('frequency', 'D')

        trading_dates = intuition.utils.build_trading_timeline(
            context.pop('start', None), context.pop('end', None),
            context['frequency'])

        context['index'] = trading_dates
        context['live'] = (dt.datetime.now(tz=pytz.utc) < trading_dates[-1])

    def _normalize_data_types(self, strategy):
        ''' some contexts only retrieves strings, giving back right type '''
        for k, v in strategy.iteritems():
            if not isinstance(v, str):
                # There is probably nothing to do
                continue
            if v == 'true':
                strategy[k] = True
            elif v == 'false' or v is None:
                strategy[k] = False
            else:
                try:
                    if v.find('.') > 0:
                        strategy[k] = float(v)
                    else:
                        strategy[k] = int(v)
                except ValueError:
                    pass

    def build(self):
        context = self.load()

        algorithm = context.pop('algorithm', {})
        manager = context.pop('manager', {})
        data = context.pop('data', {})

        if context:
            self._normalize_dates(context)

        self._normalize_data_types(algorithm)
        self._normalize_data_types(manager)
        self._normalize_data_types(data)

        strategy = {
            'algorithm': algorithm,
            'manager': manager,
            'data': data
        }

        return context, strategy



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Universe manager
  ----------------

  It knows everything about market and smartly handle user input

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import os
import random
import yaml
import dna.logging
from intuition.errors import LoadMarketSchemeFailed

log = dna.logging.logger(__name__)


class Market(object):
    '''
    Knows everything about market and smartly handle user input
    '''

    # TODO Read default from market.yml or set by the user
    benchmark = '^GSPC'
    timezone = 'US/Eastern'
    raw_description = None
    scheme_path = os.path.expanduser('~/.intuition/data/market.yml')

    def __init__(self):
        log.info('loading market scheme')
        self._load_market_scheme()

    def _load_market_scheme(self):
        ''' Load market yaml description '''
        try:
            self.scheme = yaml.load(open(self.scheme_path, 'r'))
        except Exception, error:
            raise LoadMarketSchemeFailed(reason=error)

    def _extract_forex(self):
        self.timezone = self.scheme['forex']['timezone']
        return self.scheme['forex']['pairs']

    def _extract_cac40(self, market):
        market_scheme = self.scheme
        # Walk the market structure
        for key in market[:-1]:
            market_scheme = market_scheme[key]

        self.timezone = market_scheme['timezone']
        self.benchmark = market_scheme['benchmark']
        return map(
            lambda x: x + '.pa',
            market_scheme[market[-1]].keys())

    def _lookup_sids(self, market, limit=-1):
        market = market.split(':')

        try:
            if 'forex' in market:
                sids = self._extract_forex()
            elif 'paris' in market:
                sids = self._extract_cac40(market)
            else:
                raise NotImplementedError(
                    'market {} not supported'.format(market))
        except Exception, error:
            raise LoadMarketSchemeFailed(reason=error)

        random.shuffle(sids)
        return sids[:limit] if limit > 0 else sids

    # TODO multi exchange, sector and other criterias
    def parse_universe_description(self, description):
        ''' Semantic

           - 'sid1,sid2,sid2,...'
           - 'exchange' : every sids of the exchange
           - 'exchange,n' : n random sids of the exchange

        where exchange is a combination of 'type:index:submarket'
        '''
        self.raw_description = description
        description = description.split(',')
        self.exchange = description[0]

        n = int(description[1]) if len(description) == 2 else -1
        self.sids = self._lookup_sids(description[0], n)

    def _detect_exchange(self, description):
        ''' Guess from the description and the market scheme '''
        pass

    def filter_open_hours(self, index):
        ''' Remove market specific closed hours '''
        return index

    def __str__(self):
        return '<sids: {} exchange: {} timezone: {} benchmark: {}'.format(
            self.sids, self.exchange, self.timezone, self.benchmark)


#TODO Complete with :
# http://en.wikipedia.org/wiki/List_of_stock_exchange_opening_times
# http://en.wikipedia.org/wiki/List_of_S&P_500_companies
"""
def filter_market_hours(dates, exchange):
    ''' Only return market open hours from UTC timezone'''
    #NOTE UTC times ! Beware of summer and winter hours
    if dates.freq >= pd.datetools.Day():
        # Daily or lower frequency, no hours filter required
        return dates
    if exchange == 'paris':
        selector = ((dates.hour > 6) & (dates.hour < 16)) | \
            ((dates.hour == 17) & (dates.minute < 31))
    elif exchange == 'london':
        selector = ((dates.hour > 8) & (dates.hour < 16)) | \
            ((dates.hour == 16) & (dates.minute > 31))
    elif exchange == 'tokyo':
        selector = ((dates.hour > 0) & (dates.hour < 6))
    elif exchange == 'nasdaq' or exchange == 'nyse':
        selector = ((dates.hour > 13) & (dates.hour < 21)) | \
            ((dates.hour == 13) & (dates.minute > 31))
    else:
        # Forex or Unknown market, return as is
        return dates

    # Pandas dataframe filtering mechanism
    index = dates[selector]
    if not index.size:
        raise ExchangeIsClosed(exchange=exchange, dates=dates)
    return index
"""



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition engine
  ----------------

  Wraps zipline engine and results with more configuration options

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import dna.utils
import dna.logging
from zipline.finance.trading import TradingEnvironment
from zipline.utils.factory import create_simulation_parameters
import intuition.constants as constants
from intuition.data.loader import LiveBenchmark
from intuition.core.analyzes import Analyze
import intuition.utils as utils

log = dna.logging.logger(__name__)


# NOTE Is there still a point to use here a constructor object instead of a
#     simple function ?
class TradingEngine(object):
    ''' Factory class wrapping zipline Backtester, returns the requested algo
    ready for use '''

    def __new__(self, identity, modules, strategy_conf):

        algo_obj = utils.intuition_module(modules['algorithm'])
        algo_obj.identity = identity
        trading_algo = algo_obj(properties=strategy_conf['algorithm'])

        trading_algo.set_logger(dna.logging.logger('algo.' + identity))

        # Use a portfolio manager
        if modules.get('manager'):
            log.info('initializing manager {}'.format(modules['manager']))
            # Linking to the algorithm the configured portfolio manager
            trading_algo.manager = utils.intuition_module(modules['manager'])(
                strategy_conf['manager'])
        else:
            trading_algo.manager = None
            log.info('no portfolio manager used')

        return trading_algo


class Simulation(object):
    ''' Setup and trigger trading sessions '''
    context = None

    def _get_benchmark_handler(self, last_trade, freq='minutely'):
        '''
        Setup a custom benchmark handler or let zipline manage it
        '''
        return LiveBenchmark(
            last_trade, frequency=freq).surcharge_market_data \
            if utils.is_live(last_trade) else None

    def configure_environment(self, last_trade, benchmark, timezone):
        ''' Prepare benchmark loader and trading context '''

        # Setup the trading calendar from market informations
        self.benchmark = benchmark
        self.context = TradingEnvironment(
            bm_symbol=benchmark,
            exchange_tz=timezone,
            load=self._get_benchmark_handler(last_trade))

    def build(self, identity, modules, strategy=constants.DEFAULT_CONFIG):
        '''
        Wrapper of zipline run() method. Use the configuration set so far
        to build up the trading environment
        '''
        self.engine = TradingEngine(identity, modules, strategy)
        self.initial_cash = strategy['manager'].get('cash', None)

    def __call__(self, datafeed, auto=False):
        ''' wrap zipline.run() with finer control '''
        self.engine.auto = auto
        # FIXME crash if trading one day that is not a trading day
        with self.context:
            sim_params = create_simulation_parameters(
                capital_base=self.initial_cash,
                start=datafeed.start,
                end=datafeed.end)

            daily_stats = self.engine.run(datafeed, sim_params)

        return Analyze(
            params=sim_params,
            results=daily_stats,
            metrics=self.engine.risk_report,
            benchmark=self.benchmark)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Data related utilities
  ----------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import pandas as pd


# TODO Make it a decorator
def clean_sid(sid):
    sid = str(sid).lower()
    # Remove market extension
    dot_pos = sid.find('.')
    sid = sid[:dot_pos] if dot_pos > 0 else sid
    # Remove forex slash
    return sid.replace('/', '')


def apply_mapping(raw_row, mapping):
    '''
    Override this to hand craft conversion of row.
    '''
    row = {target: mapping_func(raw_row[source_key])
           for target, (mapping_func, source_key)
           in mapping.fget().items()}
    return row


def invert_dataframe_axis(fct):
    '''
    Make dataframe index column names,
    and vice et versa
    '''
    def inner(*args, **kwargs):
        df_to_invert = fct(*args, **kwargs)
        assert isinstance(df_to_invert, pd.DataFrame)
        return pd.DataFrame(df_to_invert.to_dict().values(),
                            index=df_to_invert.to_dict().keys())
    return inner


#NOTE with_market symbol, build jxr.pa => JXR:EPA
def use_google_symbol(fct):
    '''
    Removes ".PA" or other market indicator from yahoo symbol
    convention to suit google convention
    '''
    def decorator(symbols):
        google_symbols = []

        # If one symbol string
        if isinstance(symbols, str):
            symbols = [symbols]

        symbols = sorted(symbols)
        for symbol in symbols:
            dot_pos = symbol.find('.')
            google_symbols.append(
                symbol[:dot_pos] if (dot_pos > 0) else symbol)

        data = fct(google_symbols)
        #NOTE Not very elegant
        data.columns = [s for s in symbols if s.split('.')[0] in data.columns]
        return data
    return decorator



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition misc utilities
  ------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import time
import datetime as dt
import pytz
import calendar
import locale
import dateutil.parser
import pandas as pd
import dna.utils


def is_live(current_date):
    return (current_date > pd.datetime.now(pytz.utc))


# TODO Handle in-day dates, with hours and minutes
def normalize_date_format(date):
    '''
    Dates can be defined in many ways, but zipline use
    aware datetime objects only. Plus, the software work
    with utc timezone so we convert it.
    '''
    if isinstance(date, int):
        # This is probably epoch time
        date = time.strftime('%Y-%m-%d %H:%M:%S',
                             time.localtime(date))

    # assert isinstance(date, str) or isinstance(date, unicode)
    if isinstance(date, str) or isinstance(date, unicode):
        date = dateutil.parser.parse(date)
    if not date.tzinfo:
        local_tz = pytz.timezone(_detect_timezone())
        local_dt = local_tz.localize(date, is_dst=None)
        # local_dt = local_tz.localize(parse(date), is_dst=None)
        date = local_dt.astimezone(pytz.utc)

    return date


def _detect_timezone():
    '''
    Experimental and temporary (since there is a world module)
    get timezone as set by the system
    '''
    default_timezone = 'America/New_York'
    locale_code = locale.getdefaultlocale()
    return default_timezone if not locale_code else \
        str(pytz.country_timezones[locale_code[0][-2:]][0])


def UTC_date_to_epoch(utc_date):
    return calendar.timegm(utc_date.timetuple())


def epoch_to_date(epoch, tz=pytz.utc):
    tm = time.gmtime(epoch)
    return(dt.datetime(tm.tm_year, tm.tm_mon, tm.tm_mday,
           tm.tm_hour, tm.tm_min, tm.tm_sec, 0, tz))


def next_tick(date, interval=15):
    '''
    Only return when we reach given datetime
    '''
    # Intuition works with utc dates, conversion are made for I/O
    now = dt.datetime.now(pytz.utc)
    live = False
    while now < date:
        time.sleep(interval)
        now = dt.datetime.now(pytz.utc)
        live = True
    return live


def intuition_module(location):
    ''' Build the module path and import it '''
    path = location.split('.')
    obj = path.pop(-1)
    return dna.utils.dynamic_import('.'.join(path), obj)


def truncate(float_value, n=2):
    if isinstance(float_value, float):
        float_value = float('%.*f' % (n, float_value))
    return float_value


# TODO Frequency for live trading (and backtesting ?)
def build_trading_timeline(start, end, freq='D'):
    EMPTY_DATES = pd.date_range('2000/01/01', periods=0, tz=pytz.utc)
    now = dt.datetime.now(tz=pytz.utc)

    if not start:
        if not end:
            # Live trading until the end of the day
            bt_dates = EMPTY_DATES
            live_dates = pd.date_range(
                start=now,
                end=normalize_date_format('23h'),
                freq=freq)
        else:
            end = normalize_date_format(end)
            if end < now:
                # Backtesting since a year before end
                bt_dates = pd.date_range(
                    start=end - 360 * pd.datetools.day,
                    end=end)
                live_dates = EMPTY_DATES
            elif end > now:
                # Live trading from now to end
                bt_dates = EMPTY_DATES
                live_dates = pd.date_range(start=now, end=end, freq=freq)
    else:
        start = normalize_date_format(start)
        if start < now:
            if not end:
                # Backtest for a year or until now
                end = start + 360 * pd.datetools.day
                if end > now:
                    end = now - pd.datetools.day
                live_dates = EMPTY_DATES
                bt_dates = pd.date_range(
                    start=start, end=end)
            else:
                end = normalize_date_format(end)
                if end < now:
                    # Nothing to do, backtest from start to end
                    live_dates = EMPTY_DATES
                    bt_dates = pd.date_range(start=start, end=end)
                elif end > now:
                    # Hybrid timeline, backtest from start to end, live
                    # trade from now to end
                    bt_dates = pd.date_range(
                        start=start, end=now - pd.datetools.day)

                    live_dates = pd.date_range(
                        start=now, end=end, freq=freq)
        elif start > now:
            if not end:
                # Live trading from start to the end of the day
                bt_dates = EMPTY_DATES
                live_dates = pd.date_range(
                    start=start,
                    end=normalize_date_format('23h'),
                    freq=freq)
            else:
                # Live trading from start to end
                end = normalize_date_format(end)
                bt_dates = EMPTY_DATES
                live_dates = pd.date_range(
                    start=start,
                    end=end, freq=freq)

    return bt_dates + live_dates



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition remote data
  ---------------------

  Provides functions to access various trading data accross internet wires

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache2.0, see LICENSE for more details.
'''

from pandas.rpy.common import convert_to_r_matrix
import requests
from pandas.io.data import DataReader, get_quote_yahoo
import pandas as pd
import json
from xml.dom import minidom, Node
import urllib2
import dna.logging
from zipline.utils.factory import load_from_yahoo, load_bars_from_yahoo
from intuition.data.utils import (
    use_google_symbol, invert_dataframe_axis, apply_mapping)

log = dna.logging.logger(__name__)

finance_urls = {
    'yahoo_hist': 'http://ichart.yahoo.com/table.csv',
    'yahoo_infos': 'http://finance.yahoo/q/pr',
    'google_prices': 'http://www.google.com/finance/getprices',
    'snapshot_google_light': 'http://www.google.com/finance/info',
    'snapshot_google_heavy': 'http://www.google.com/ig/api',
    'info_lookup': ('http://d.yimg.com/autoc.finance.yahoo.com/autoc?' +
                    'query={}&callback=YAHOO.Finance.SymbolSuggest.ssCallback')
}


class Data(object):
    '''
    Entry point to remote access to data
    Mainly offer a simpler interface and a data access harmonisation
    Plus symbol check and complete, timezone, currency conversion
    Plus reindexage ?
    '''
    def __init__(self, country_code=None):
        '''
        Parameters
            country_code: str
                This information is used to setup International object
                and get the right local conventions for lang,
                dates and currencies. None will stand for 'fr'
        '''
        #self.locatioon = world.International(country_code)

    def fetch_equities_daily(self, symbols, ohlc=False,
                             r_type=False, returns=False, **kwargs):
        if len(symbols) == 0:
            return pd.DataFrame()
        if isinstance(symbols, str):
            symbols = symbols.split(',')

        if ohlc:
            data = load_bars_from_yahoo(stocks=symbols, **kwargs)
            #data.items = symbols
            data = data.dropna()
        else:
            data = load_from_yahoo(stocks=symbols, **kwargs)
            #data.columns = symbols

            #NOTE Would it work with a pandas panel ?
            if returns:
                data = ((data - data.shift(1)) / data).fillna(method='bfill')
            if r_type:
                data = convert_to_r_matrix(data)

        return data

    def fetch_equities_snapshot(self, *args, **kwargs):
        '''
        Use Yahoo and google finance service to fetch
        current infromations about given equitiy names
        ______________________________________________
        Parameters
            args: tuple | list
                company symbols to consider
            kwargs['level']: int
                Quantity of information level
        ______________________________________________
        Return
            snapshot: pandas.DataFrame
                with names as columns and informations as index
        '''
        symbols = args
        if not symbols:
            symbols = kwargs.get('symbols', [])
        level = kwargs.get('level', 0)
        #TODO Symbols are usualy reused, cach them

        if not level:
            # default level, the lightest
            snapshot = snapshot_yahoo_pandas(symbols)
        elif level == 1:
            snapshot = snapshot_google_light(symbols)
        elif level == 2:
            snapshot = snapshot_google_heavy(symbols)
        else:
            raise ValueError('Invalid level of information requested')

        # Give columns back equities names requested
        #snapshot.columns = equities
        return snapshot

    def _localize_data(self, data):
        '''
        Inspect data and apply localization on dates and monnaie values
        '''
        assert isinstance(data, pd.Dataframe) or isinstance(data, dict)


#NOTE check zipline.utils.factory, use DataReader as well
def historical_pandas_yahoo(symbol, source='yahoo', start=None, end=None):
    '''
    Fetch from yahoo! finance historical quotes
    '''
    #NOTE Panel for multiple symbols ?
    #NOTE Adj Close column name not cool (a space)
    return DataReader(symbol, source, start=start, end=end)


#NOTE From here every methods has the same signature:
#     pandas.DataFrame = fct(yahoo_symbol(s))
# Index symbol ex: ^fchi
@invert_dataframe_axis
def snapshot_yahoo_pandas(symbols):
    '''
    Get a simple snapshot from yahoo, return dataframe
    __________________________________________________
    Return
        pandas.DataFrame with symbols as index
        and columns = [change_pct, time, last, short_ratio, PE]
    '''
    if isinstance(symbols, str):
        symbols = [symbols]
    # TODO lower() columns
    data = get_quote_yahoo(symbols)
    data.columns = map(str.lower, data.columns)
    return data


#NOTE Can use symbol with market: 'goog:nasdaq', any difference ?
@use_google_symbol
def snapshot_google_light(symbols):
    payload = {'client': 'ig', 'q': ','.join(symbols)}
    response = requests.get(finance_urls['snapshot_google_light'],
                            params=payload)
    try:
        json_infos = json.loads(response.text[3:], encoding='utf-8')
    except ValueError:
        log.error("no reliable data found")
        return pd.DataFrame()

    snapshot = {}
    for i, quote in enumerate(json_infos):
        #FIXME nasdaq and nyse `symbols` are in capital, not cac40
        if quote['t'].lower() in map(str.lower, symbols):
            #snapshot[symbols[i]] = intuition.utils.apply_mapping(
            snapshot[symbols[i]] = apply_mapping(
                quote, google_light_mapping)
        else:
            log.warning('Unknown symbol {}, ignoring...'.format(
                quote['t']))

    return pd.DataFrame(snapshot)


#TODO all values are string, make them floats (with a mapping)
#FIXME It seems this is deprecated Oo
@use_google_symbol
def snapshot_google_heavy(symbols):
    url = (finance_urls['snapshot_google_heavy'] +
           '?stock=' + '&stock='.join(symbols))
    log.info('Retrieving heavy Snapshot from %s' % url)
    try:
        url_fd = urllib2.urlopen(url)
    except IOError:
        log.error('** Bad url: %s' % url)
        return pd.DataFrame()

    try:
        xml_doc = minidom.parse(url_fd)
        root_node = xml_doc.documentElement
    except:
        log.error('** Parsing xml google response')
        return pd.DataFrame()
    i = 0
    snapshot = {q: {} for q in symbols}
    for i, node in enumerate(root_node.childNodes):
        if (node.nodeName != 'finance'):
            continue
        for item_node in node.childNodes:
            if (item_node.nodeType != Node.ELEMENT_NODE):
                continue
            snapshot[symbols[i]][item_node.nodeName] = \
                item_node.getAttribute('data')

    return pd.DataFrame(snapshot)


@property
def google_light_mapping():
    return {
        'change': (str, 'c'),
        'str_change': (str, 'ccol'),
        'perc_change': (float, 'cp'),
        'exchange': (str, 'e'),
        'id': (int, 'id'),
        'price': (str, 'l'),
        'last_price': (lambda x: x, 'l_cur'),
        'date': (str, 'lt'),
        'time': (str, 'ltt'),
        's': (int, 's'),
        'symbol': (str, 't'),
    }


def lookup_symbol(company):
    ''' Fetch company info
    >>> lookup_symbol("Apple Inc.")
          {u'exch': u'NMS',
           u'market': u'NASDAQ',
           u'name': u'Apple Inc.',
           u'symbol': u'AAPL',
           u'type': u'Equity'}
    '''
    request = requests.get(finance_urls['info_lookup'].format(company))
    return json.loads(request.text[39:-1])["ResultSet"]["Result"] \
        if request.ok else {'error': request.reason}



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition Data source api
  -------------------------

  Trading data source factory. Everything that can extend in a useful way
  zipline.DataSource goes here. Future data sources should inherit from
  this class.

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


#import abc
import pandas as pd
import dna.logging
from zipline.sources.data_source import DataSource
from zipline.gens.utils import hash_args
import intuition.utils as utils
from intuition.errors import LoadDataFailed


def _build_event(date, sid, series):
    ''' Format the input to a zipline compliant event '''
    event = {
        'dt': date,
        'sid': sid}
    if isinstance(series, pd.Series):
        if sid in series.keys():
            event.update({'price': series[sid],
                          'volume': 1000})
        else:
            event.update(series.to_dict())
    elif isinstance(series, pd.DataFrame):
        event.update(series[sid].to_dict())

    return event


class HybridDataFactory(DataSource):
    '''
    Surcharge of zipline.DataSource, switching automatically between live
    stream and backtest sources
    '''

    backtest = None
    live = None
    _is_live = False

    def __init__(self, **kwargs):
        self.log = dna.logging.logger(__name__)

        assert isinstance(kwargs.get('index'),
                          pd.tseries.index.DatetimeIndex)

        # Unpack config dictionary with default values.
        self.sids = kwargs.get('sids') or kwargs.get('universe').sids
        self.start = kwargs.get('start', kwargs['index'][0])
        self.end = kwargs.get('end', kwargs['index'][-1])
        self.index = kwargs['index']

        # Hash_value for downstream sorting.
        self.arg_string = hash_args(**kwargs)
        self._raw_data = None
        # TODO Fails if the class as no get_data method
        if 'backtest' in kwargs:
            self.backtest = kwargs['backtest'](self.sids, kwargs)
        if 'live' in kwargs:
            self.live = kwargs['live'](self.sids, kwargs)

    @property
    def mapping(self):
        if self._is_live:
            return self.live.mapping
        else:
            return self.backtest.mapping

    def raw_data_gen(self):
        # The first date is usually a few seconds before now,
        # so we compare to the next one
        if self.backtest and not utils.is_live(self.index[1]):
            try:
                bt_data = self.backtest.get_data(
                    self.sids, self.start, self.end)
            except Exception as error:
                raise LoadDataFailed(sids=self.sids, reason=error)
        else:
            bt_data = None

        for date in self.index:
            self.log.debug('--> next tick {}'.format(date))
            self._is_live = utils.next_tick(date)
            if not self._is_live:
                date = date.replace(hour=0)
                if isinstance(bt_data, pd.DataFrame):
                    if date not in bt_data.index:
                        continue
                    for sid in self.sids:
                        series = bt_data.ix[date]
                        yield _build_event(date, sid, series)

                elif isinstance(bt_data, pd.Panel):
                    if date not in bt_data.major_axis:
                        continue
                    df = bt_data.major_xs(date)
                    for sid, series in df.iterkv():
                        yield _build_event(date, sid, series)
            else:
                try:
                    snapshot = self.live.get_data(self.sids)
                except Exception as error:
                    raise LoadDataFailed(sids=self.sids, reason=error)
                for sid, series in snapshot.iterkv():
                    yield _build_event(date, sid, series)

    @property
    def instance_hash(self):
        return self.arg_string

    @property
    def raw_data(self):
        if not self._raw_data:
            self._raw_data = self.raw_data_gen()
        return self._raw_data



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Quandl.com access
  -----------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache2.0, see LICENSE for more details.
'''


import pandas as pd
import pytz
import os
import dna.logging
import Quandl
import intuition.data.utils as utils

log = dna.logging.logger(__name__)


def _build_quandl_code(symbol):
    dot_pos = symbol.find('.')
    slash_pos = symbol.find('/')
    if dot_pos > 0:
        market = symbol[dot_pos + 1:]
        provider = 'YAHOO'
        symbol = symbol[:dot_pos]
        code = '{}_{}'.format(market, symbol)
    else:
        if slash_pos > 0:
            pair = symbol.split('/')
            provider = 'QUANDL'
            code = '{}{}'.format(pair[0], pair[1])
        else:
            market = 'NASDAQ'
            provider = 'GOOG'
            code = '{}_{}'.format(market, symbol)
    return '{}/{}'.format(provider, code).upper()


def use_quandl_symbols(fct):
    def decorator(self, symbols, **kwargs):
        if not isinstance(symbols, list):
            symbols = [symbols]
        quandl_symbols = map(_build_quandl_code, symbols)
        raw_data = fct(self, quandl_symbols, **kwargs)

        data = {}
        for sid in symbols:
            data[sid] = raw_data.filter(regex='.*{}.*'.format(
                utils.clean_sid(sid).upper()))
            data[sid].columns = map(
                lambda x: x.replace(' ', '_').lower().split('_-_')[-1],
                data[sid].columns)
        return data
    return decorator


def fractionate_request(fct):
    def inner(self, symbols, **kwargs):
        tolerance_window = 10
        cursor = 0
        data = {}
        while cursor < len(symbols):
            if cursor + tolerance_window > len(symbols):
                limit = len(symbols)
            else:
                limit = cursor + tolerance_window
            symbols_fraction = symbols[cursor:limit]
            cursor += tolerance_window
            data_fraction = fct(self, symbols_fraction, **kwargs)
            for sid, df in data_fraction.iteritems():
                data[sid] = df
        return pd.Panel(data).fillna(method='pad')
    return inner


class DataQuandl(object):
    '''
    Quandl.com as datasource
    '''
    def __init__(self, quandl_key=None):
        self.quandl_key = quandl_key if quandl_key \
            else os.environ.get('QUANDL_API_KEY', None)

    @fractionate_request
    @use_quandl_symbols
    def fetch(self, code, **kwargs):
        '''
        Quandl entry point in datafeed object
        '''
        log.debug('fetching QuanDL data (%s)' % code)
        # This way you can use your credentials even if
        # you didn't provide them to the constructor
        if 'authtoken' in kwargs:
            self.quandl_key = kwargs.pop('authtoken')

        # Harmonization: Quandl call start trim_start
        if 'start' in kwargs:
            kwargs['trim_start'] = kwargs.pop('start')
        if 'end' in kwargs:
            kwargs['trim_end'] = kwargs.pop('end')

        try:
            data = Quandl.get(code, authtoken=self.quandl_key, **kwargs)
            # FIXME With a symbol not found, insert a not_found column
            data.index = data.index.tz_localize(pytz.utc)
            #data.columns = map(
            #   lambda x: x.replace(' ', '_').lower(), data.columns)
        except:
            log.error('** unable to fetch %s from Quandl' % code)
            data = pd.DataFrame()
        return data



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition alogirthm api
  -----------------------

  Trading algorithms factory. Everything that can extend in a useful way
  zipline.TradingAlgorithm goes here. Future algorithms should inherit from
  this class.

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import abc
import datetime as dt
from zipline.algorithm import TradingAlgorithm
import zipline.finance.commission as commission
from intuition.errors import AlgorithmEventFailed
import insights.plugins.database as database
import insights.plugins.mobile as mobile
import insights.plugins.hipchat as hipchat
import insights.plugins.messaging as msg


class TradingFactory(TradingAlgorithm):
    '''
    Intuition surcharge of main zipline class, but fully compatible.
    Its role is slightly different : the user will eventually expose an event()
    method meant to return buy and sell signals, processed further by the
    portfolio strategy. However it remains fully compatible with zipline method
    (i.e. it's stil possible to execute orders within the event method and not
    consider any portfolio strategy aside)
    '''

    __metaclass__ = abc.ABCMeta

    #TODO Use another zipline mechanism
    days = 0
    sids = []
    middlewares = []
    auto = False

    def __init__(self, *args, **kwargs):
        self.realworld = kwargs['properties'].get('realworld')
        TradingAlgorithm.__init__(self, *args, **kwargs)

    def _is_interactive(self):
        ''' Prevent middlewares and orders to work outside live mode '''
        return not (
            self.realworld and (dt.date.today() > self.datetime.date()))

    # FIXME It needs insights to be installed
    def use_default_middlewares(self, properties):
        if properties.get('interactive'):
            self.use(msg.RedisProtocol(self.identity).check)
        device = properties.get('mobile')
        if device:
            self.use(mobile.AndroidPush(device).notify)
        if properties.get('save'):
            self.use(database.RethinkdbFinance(
                table=self.identity, db='portfolios', reset=True)
                .save_portfolio)
        hipchat_room = properties.get('hipchat')
        if hipchat_room:
            self.use(hipchat.Bot(
                hipchat_room, name=self.identity).notify)

        self.set_commission(commission.PerTrade(
            cost=properties.get('commission', 2.5)))

    def use(self, func, when='whenever'):
        ''' Append a middleware to the algorithm '''
        #NOTE A middleware Object ?
        print('registering middleware {}'.format(func.__name__))
        self.middlewares.append({
            'call': func,
            'name': func.__name__,
            'args': func.func_code.co_varnames,
            'when': when})

    #NOTE I'm not superfan of initialize + warm
    def warm(self, data):
        ''' Called at the first handle_data frame '''
        pass

    @abc.abstractmethod
    def event(self, data):
        ''' User should overwrite this method '''
        pass

    def handle_data(self, data):
        ''' Method called for each event by zipline. In intuition this is the
        place to factorize algorithms and then call event() '''
        self.days += 1
        signals = {}
        self.orderbook = {}

        if self.initialized and self.manager:
            self.manager.update(
                self.portfolio,
                self.datetime,
                self.perf_tracker.cumulative_risk_metrics.to_dict())
        else:
            # Perf_tracker needs at least a turn to have an index
            self.sids = data.keys()
            self.warm(data)
            self.initialized = True
            return

        try:
            signals = self.event(data)
        except Exception, error:
            # NOTE Temporary debug. Will probably notify the error and go on
            # with signals=None
            raise AlgorithmEventFailed(
                reason=error, date=self.datetime, data=data)

        # One can process orders within the alogrithm and don't return anything
        if signals and self.manager:
            if (signals.get('buy') or signals.get('sell')):
                self.orderbook = self.manager.trade_signals_handler(signals)
                if self.auto and self._is_interactive():
                    self.process_orders(self.orderbook)

        if self._is_interactive():
            self._call_middlewares()

    def process_orders(self, orderbook):
        ''' Default and costant orders processor. Overwrite it for more
        sophisticated strategies '''
        for stock, alloc in orderbook.iteritems():
            self.logger.info('{}: Ordered {} {} stocks'.format(
                self.datetime, stock, alloc))
            if isinstance(alloc, int):
                self.order(stock, alloc)
            elif isinstance(alloc, float) and \
                    alloc >= -1 and alloc <= 1:
                self.order_percent(stock, alloc)
            else:
                self.logger.warning(
                    '{}: invalid order for {}: {})'
                    .format(self.datetime, stock, alloc))

    def _call_one_middleware(self, middleware):
        ''' Evaluate arguments and execute the middleware function '''
        args = {}
        for arg in middleware['args']:
            if hasattr(self, arg):
                # same as eval() but safer for arbitrary code execution
                args[arg] = reduce(getattr, arg.split('.'), self)
        self.logger.debug('calling middleware event {}'
                          .format(middleware['name']))
        middleware['call'](**args)

    def _check_condition(self, when):
        ''' Verify that the middleware condition is True '''
        #TODO Use self.datetime to evaluate <when>
        return True

    def _call_middlewares(self):
        ''' Execute the middleware stack '''
        for middleware in self.middlewares:
            if self._check_condition(middleware['when']):
                self._call_one_middleware(middleware)
